////////////////////////////
// Server side of web-interface FACS Plot
//
// by Mark Hammer
////////////////////////////

package FACS.Plot.PlotServlet;

import FACS.*;
import FACS.Plot.Region;
import FACS.Plot.RVector;
import javax.servlet.*;
import javax.servlet.http.*;
import javax.naming.Context;
import java.io.*;
import java.util.*;
import java.nio.ByteBuffer;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.regex.*;

import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import java.awt.geom.*;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.Polygon;
import java.nio.channels.FileChannel;
import java.nio.MappedByteBuffer;
import java.beans.XMLEncoder;
import java.beans.XMLDecoder;

public class PlotServlet extends HttpServlet {
	private static String ROOT;

	public static int CACHE_SIZE = 20; //max number of FCS files in memory at once

	public static final java.beans.ExceptionListener el = new java.beans.ExceptionListener() {
			public void exceptionThrown(Exception exception) {
				exception.printStackTrace();
			}
		};
		
	private ReentrantReadWriteLock rw = new ReentrantReadWriteLock();

	static final int DOTPLOT = 1;
    static final int HIST = 2;
    static final int SM_HIST = 3;
    static final int DENSITY = 4;
    static final int CONTOUR = 5;

	static final int NONE = 0;
    static final int ELLIPSE = 1;
    static final int RECT = 2;
    static final int POLY = 3;
    static final int QUAD = 4;
    static final int RANGE = 5;

	public static TimedConcurrentHashMap regionsCache, compCache, dataCache;

    public PlotServlet() {
		super();

		try {
			Context c = new javax.naming.InitialContext();
			ROOT = (String)c.lookup("java:comp/env/fcsroot");
		} catch (javax.naming.NamingException ne) {
			ne.printStackTrace();
			ROOT = "";
		}
		
		regionsCache = new TimedConcurrentHashMap(20);
		compCache = new TimedConcurrentHashMap(20);
		dataCache = new TimedConcurrentHashMap(CACHE_SIZE);
    }

	public void doGet(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
		doPost(request,response);
	}

	public void doPost(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
		/*HttpSession session = request.getSession();
		
		//check that we haven't changed directory
		File tempdir = new File(ROOT,request.getParameter("directory"));
		rw.readLock().lock();
		File directory = (File)session.getAttribute("directory");
		File filedirectory = (File)session.getAttribute("filedirectory");
		if(directory == null || !tempdir.equals(directory) || session.getAttribute("regions") == null) {
			rw.readLock().unlock();
			rw.writeLock().lock();
			directory = tempdir;
			filedirectory = new File(ROOT,XMLFiles.getFileLink(ROOT,request.getParameter("directory")));
			session.setAttribute("directory",directory);
			session.setAttribute("filedirectory",filedirectory);
			initialize(session);
			rw.writeLock().unlock();
			rw.readLock().lock();
		} else {
			if(isNewCompMatrix(session)) { //isNewCompMatrix will actually load the new one...
				rw.readLock().unlock();
				rw.writeLock().lock();
				session.setAttribute("data",new TimedConcurrentHashMap(CACHE_SIZE));
				rw.writeLock().unlock();
				rw.readLock().lock();
			}
		}
		rw.readLock().unlock();*/
		
		File directory = new File(ROOT,request.getParameter("directory"));
		File filedirectory = new File(ROOT,XMLFiles.getFileLink(ROOT,request.getParameter("directory")));
		
		rw.readLock().lock();
		if(regionsCache.get(directory) == null) { // || compCache.get(directory) == null //not sure why this was here, but it breaks stuff that doesn't have a comp matrix
			rw.readLock().unlock();
			rw.writeLock().lock();
			if(regionsCache.get(directory) == null) { // || compCache.get(directory) == null
				initialize(directory);
			}
			rw.writeLock().unlock();
			rw.readLock().lock();
		} else if(isNewCompMatrix(directory)) {
			rw.readLock().unlock();
			rw.writeLock().lock();
			clearCache(directory);
			rw.writeLock().unlock();
			rw.readLock().lock();
		}
		rw.readLock().unlock();
		
		

		String mode = request.getParameter("mode");
		if(mode.equals("regions")) {
			getRegions(request,directory);
		} else if(mode.equals("plot")) {
			doPlot(request,response,directory,filedirectory);
		}
	}

	public void getRegions(HttpServletRequest request, File directory) throws ServletException, IOException {
		rw.writeLock().lock();
		
		//gets regions from client and saves them
		RVector regions = (RVector)regionsCache.get(directory);
		RVector oldRegions = (RVector)regions.clone();
		regions = new RVector();

		BufferedInputStream is = new BufferedInputStream(request.getInputStream());
		ObjectInputStream ois = new ObjectInputStream(is);
		try {
			regions = (RVector)ois.readObject();
			ois.close();
		} catch (ClassNotFoundException cnfe) {
			cnfe.printStackTrace();
			return;
		}
		
		//session.setAttribute("regions",regions);
		regionsCache.put(directory,regions);
		rw.writeLock().unlock();
		
		//only save if asked to, else we're done
		if(!request.getParameter("save").equalsIgnoreCase("true"))
			return;
		
		
		rw.writeLock().lock(); //because we have to write to regions.xml...
		File xml;

		//compare old and new regions... add changed regions to rchanged.xml
		Hashtable rchanged;
		try {
			xml = new File(directory,"rchanged.xml");
			XMLDecoder d = new XMLDecoder(new BufferedInputStream(new FileInputStream(xml)));
			d.setExceptionListener(el);
			rchanged = (Hashtable)d.readObject();
		} catch (FileNotFoundException nfe) {
			rchanged = new Hashtable();
		}
		for(int i = 0; i < regions.size(); i++) { //don't care about new or deleted regions
												//(at present, we assume they won't delete important regions...
												//since then we don't really know what to do)
			Region newR = (Region)regions.elementAt(i);
			Region oldR = getRegionByName(oldRegions, newR.getName());
			if(oldR == null) //region is completely new
				continue;
			HashSet diff = getDifferences(newR,oldR);
			if(diff.size() == 0) //no differences, done.
				continue;
			if(rchanged.containsKey(newR.getName())) {
				((HashSet)rchanged.get(newR.getName())).addAll(diff);
			} else {
				rchanged.put(newR.getName(),diff);
			}
		}
		xml = new File(directory,"rchanged.xml");
		xml.delete();
		if(!rchanged.isEmpty()) {
			FileOutputStream fos = new FileOutputStream(xml);
			XMLEncoder xenc = new XMLEncoder(new BufferedOutputStream(new FileOutputStream(xml)));
			xenc.setExceptionListener(el);
			xenc.writeObject(rchanged);
			xenc.flush();
			xenc.close();
			fos.close();
		}

		//save new regions to regions.xml
		xml = new File(directory,"regions.xml");
		xml.delete();
		FileOutputStream fos = new FileOutputStream(xml);
		XMLEncoder xenc = new XMLEncoder(new BufferedOutputStream(new FileOutputStream(xml)));
		xenc.setExceptionListener(el);
		for(int i = 0; i < regions.size(); i++) {
			xenc.writeObject((Region)regions.elementAt(i));
			xenc.flush();
		}
		xenc.close();
		fos.close();
		rw.writeLock().unlock();
	}

	public Region getRegionByName(Vector regions, String name) {
		for(int j = 0; j < regions.size(); j++) {
			if(((Region)regions.elementAt(j)).getName().equals(name)) {
				return (Region)regions.elementAt(j);
			}
		}
		return null;
	}

	public HashSet getDifferences(Region one, Region two) {
		//we want differences in either direction
		HashSet out = new HashSet();
		HashSet totalKeys = new HashSet(one.getValues().keySet());
		totalKeys.addAll(two.getValues().keySet());
		//for(int i = 0; i < totalKeys.size(); i++) {
		for(Iterator i = totalKeys.iterator(); i.hasNext(); ) {
			//String file = (String)totalKeys.elementAt(i);
			String file = (String)i.next();
			double[] oneV = one.getVals(file);
			double[] twoV = two.getVals(file);
			if(!Arrays.equals(oneV,twoV))
				out.add(file);
		}
		return out;
	}

	public void doPlot(HttpServletRequest request,
			HttpServletResponse response, File directory, File filedirectory) throws ServletException, IOException {
		
		int param1, param2, mode;
		int resX, resY;
		boolean hSecondMax, cShowOutliers;
		boolean smooth; //option for density plot
		boolean comp;
		int nEvents = -1; //maximum number of events to load (default 50k on client side; -1 = all)
		
		if(request.getParameter("comp") != null) {
			if(request.getParameter("comp").equals("true"))
				comp = true;
			else
				comp = false;
		} else {
			comp = true;
		}
		File xml = new File(directory, "comp.xml");
		if(!xml.exists()) {
			comp = false;
		}
		
		if(request.getParameter("nevents") != null)
			nEvents = new Integer(request.getParameter("nevents")).intValue();

		String[] files = request.getParameter("file").split(";");
		//int[] indices = new int[files.length]; // paintPlot takes indices
												// (legacy)
		
		/*checkVectorSizes(); // only do once before, otherwise may screw up
							// PlotApplet's data indices
							 * 
							 */
		
		Vector fData = new Vector();
		
		for(int i = 0; i < files.length; i++) {
			/*File f = new File(filedirectory, files[i]);
			int in = getFDIndex(files[i], comp);
			if(in == -1) { // if we haven't already loaded it
				FACSData fd;
				try {
					fd = readFile(fData.size(), f, comp, (double[][])session.getAttribute("c"));
				} catch (ReadException re) {
					continue;
				}
				indices[i] = fData.size();
				fData.add(fd);
				fGated.add(fd);
			} else {
				indices[i] = in;
			}*/
			
			FCSFile hashkey = new FCSFile(new File(directory, files[i]),comp); //note that I put the directory as directory, not filedirectory
											//this is so that we can have different comp matrices in the duplicated experiments
			
			rw.readLock().lock();
			FACSData fd = (FACSData)dataCache.get(hashkey);
			if(fd == null) {
				rw.readLock().unlock();
				rw.writeLock().lock();
				fd = (FACSData)dataCache.get(hashkey);
				if(fd == null) { //just so we don't have a queue waiting to load this file multiple times
					File f = new File(filedirectory, files[i]);
					try {
						fd = readFile(fData.size(), f, nEvents, comp, (double[][])compCache.get(directory));
					} catch (ReadException re) {
						re.printStackTrace();
						continue;
					}
					dataCache.put(hashkey,fd);
				}
				rw.writeLock().unlock();
				rw.readLock().lock();
			}
			fData.add(fd);
			rw.readLock().unlock();
		}

		param1 = new Integer(request.getParameter("param1")).intValue();
		param2 = new Integer(request.getParameter("param2")).intValue();
		mode = new Integer(request.getParameter("pmode")).intValue();
		if(mode == DENSITY || mode == CONTOUR || mode == HIST || mode == SM_HIST) {
			resX = 256;
			resY = 256;
		} else {
			resX = 300;
			resY = 300;
		}
		
		hSecondMax = new Boolean(request.getParameter("hSecondMax")).booleanValue();
		smooth = false;
		if(request.getParameter("smooth") != null)
			smooth = new Boolean(request.getParameter("smooth")).booleanValue();
		cShowOutliers = false;
		if(request.getParameter("outliers") != null)
			cShowOutliers = new Boolean(request.getParameter("outliers")).booleanValue();

		//Gate data
		String[] gatesS = request.getParameter("gates").split("\n");
		//String[] pcs = gatesS[0].split("\\|");
		/*int[] gindices = new int[pcs.length];
		for(int i = 0; i < pcs.length; i++) {
			gindices[i] = getFDIndex(pcs[i],comp);
		}*/
		Vector gates = new Vector();
		String[] pcs2;
		for(int i = 1; i < gatesS.length; i++) {
			pcs2 = gatesS[i].split("\t");
			gates.add(Gate.fromWritable(pcs2[1],pcs2[0]));
		}
		//if(request.getParameter("gatesChanged").equals("true"))
		gate((RVector)regionsCache.get(directory),fData,gates);
		
		
		response.setContentType("image/jpeg");
		BufferedImage bi = new BufferedImage(resX,resY,BufferedImage.TYPE_4BYTE_ABGR);
		Graphics2D g2 = (Graphics2D)bi.getGraphics();
		paintPlot(g2,fData,param1,param2,mode, resX, resY, hSecondMax, cShowOutliers, smooth);
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		ImageIO.write(bi,"jpg",baos);
		response.setContentLength(baos.size());
		BufferedOutputStream bos = new BufferedOutputStream(response.getOutputStream());
		bos.write(baos.toByteArray());
		bos.close();
		baos.close();
		
		//clean up for memory...
		for(int i = 0; i < fData.size(); i++) {
			FACSData fd = (FACSData)fData.elementAt(i);
			fd.data = null;
			fd = null;
		}
		fData = null;
		bi = null;
		System.gc();
	}

	public boolean initialize(File directory) {
		//get regions
		RVector regions = new RVector();
		File xml = new File(directory,"regions.xml");
		if(xml.exists()) {
			try {
				XMLDecoder d = new XMLDecoder(new BufferedInputStream(new FileInputStream(xml)));
				d.setExceptionListener(el);
				while(true) {
					try {
						regions.add((Region)d.readObject());
					} catch(ArrayIndexOutOfBoundsException aioobe) { //file is read completely
						d.close();
						break;
					}
				}
			} catch (IOException ioe) {
				ioe.printStackTrace();
			}
		}
		
		//session.setAttribute("regions",regions);
		regionsCache.put(directory,regions);
		
		//load compensation matrix
		xml = new File(directory,"comp.xml");
		boolean comp = true;
		if(xml.exists()) {
			try {
				XMLDecoder d = new XMLDecoder(new BufferedInputStream(new FileInputStream(xml)));
				d.setExceptionListener(el);
				try {
					double[][] c = (double[][])d.readObject();
					//session.setAttribute("c",c);
					compCache.put(directory,c);
				} finally { //file is read completely
					d.close();
				}
			} catch (IOException ioe) {
				ioe.printStackTrace();
				comp = false;
			}
		} else {
			comp = false;
		}
		return comp;
	}

	/*public void checkVectorSizes() {
		if(fData == null)
			fData = new Vector();
		if(fGated == null)
			fGated = new Vector();
		while(fData.size() > MAX_FILES) {
			fData.remove(0);
			fGated.remove(0);
		}
	}*/
	
	public boolean isNewCompMatrix(File directory) {
		File xml = new File(directory,"comp.xml");
		double[][] c = (double[][])compCache.get(directory);
		if(xml.exists()) {
			try {
				XMLDecoder d = new XMLDecoder(new BufferedInputStream(new FileInputStream(xml)));
				d.setExceptionListener(el);
				double[][] c1;
				try {
					c1 = (double[][])d.readObject();
				} finally { //file is read completely
					d.close();
				}
				boolean out;
				if(c == null)
					out = true;
				else
					out = !arraysEqual(c,c1);
				c = c1;
				compCache.put(directory,c);
				return out;
			} catch (IOException ioe) {
				ioe.printStackTrace();
				//comp = false;
				return false;
			}
		} else {
			//comp = false;
			if(c == null)
				return false;
			compCache.put(directory,null);
			return true;
		}
	}
	
	public void clearCache(File directory) {
		Enumeration files = dataCache.keys();
		FCSFile temp;
		while(files.hasMoreElements()) {
			temp = (FCSFile)files.nextElement();
			if(temp.isInDirectory(directory))
				dataCache.remove(temp);
		}
	}
	
	public boolean arraysEqual(double[][] a, double[][] b) {
		double MAX_ERROR = 0.0005; //i.e. 0.05% (display value)
		if(a.length != b.length || a[0].length != b[0].length)
			return false;
		double error;
		for(int i = 0; i < a.length; i++) {
			for(int j = 0; j < a[i].length; j++) {
				error = Math.abs(a[i][j]-b[i][j]);
				if(error > MAX_ERROR)
					return false;
			}
		}
		return true;
	}

	/*public int getFDIndex(FACSData fd) {
		return getFDIndex(fd.filename,fd.comp);
	}

	public int getFDIndex(String fd, boolean comp) {
		for(int i = 0; i < fData.size(); i++) {
			if(fd.equals(((FACSData)fData.elementAt(i)).filename)) {
				if(((FACSData)fData.elementAt(i)).comp == comp)
					return i;
			}
		}
		return -1;
	}*/

	public FACSData readFile(int dataIndex, File f, int nEvents, boolean comp, double[][] c) throws ReadException {
		/*if(c == null)
			comp = false;*/
        try {
			//System.out.println(System.currentTimeMillis());
            FileInputStream fis = new FileInputStream(f);
			
            
            //file type specified "FCSn.0"
            byte[] six = new byte[6];
            fis.read(six);
            String s = new String(six);
            if(!s.equals("FCS1.0") && !s.equals("FCS2.0") && !s.equals("FCS3.0"))
                throw new ReadException("File Format Unrecognized\n(says "+s+")");
            
            //spaces
            fis.skip(4);
            
            //starts and ends of TEXT, DATA, ANALYSIS
            byte[] eight = new byte[8];
            fis.read(eight);
            int txtStart = getInt(eight);
            fis.read(eight);
            int txtEnd = getInt(eight);
            fis.read(eight);
            int datStart = getInt(eight);
            fis.read(eight);
            int datEnd = getInt(eight);
            //don't care about analysis
            
            if(txtStart == 0 || txtEnd == 0)
                throw new ReadException("FCS3.0 Long Files Unsupported If TEXT Section is long");
            
            String txt; //byte[] data = null;
			ByteBuffer bb; FileChannel fc; MappedByteBuffer mbb;
			if(datStart == 0) {
            	fis.skip(txtStart - (6+4+8+8+8+8));
                
                byte[] text = new byte[txtEnd-txtStart+1];
                fis.read(text);
                txt = new String(text);
                
                //need to get data start & end from text...
                String del = txt.substring(0,1); //delimiter
                Matcher non = Pattern.compile("[^A-Za-z]").matcher(del);
                String del1 = del;
                if(non.find())
                    del1 = "\\"+del;
                String[] list = txt.split(del1);
                Pattern patBegin = Pattern.compile("BEGINDATA");
                Pattern patEnd = Pattern.compile("ENDDATA");
                for(int i = 0; i < list.length; i++) {
                	if(patBegin.matcher(list[i]).find()) {
                		datStart = new Integer(list[i+1].trim()).intValue();
                	} else if(patEnd.matcher(list[i]).find()) {
                		datEnd = new Integer(list[i+1].trim()).intValue();
                	}
                }
                
                if(datStart == 0 || datEnd == 0)
                	throw new ReadException("Error in FCS3 Long File Data Begin/End Keys");
                
                
            } else if(txtStart < datStart) {
                fis.skip(txtStart - (6+4+8+8+8+8));
                
                byte[] text = new byte[txtEnd-txtStart+1];
                fis.read(text);
                txt = new String(text);
                
                /*fis.skip(datStart-txtEnd-1);
                
                data = new byte[datEnd-datStart+1];

                fis.read(data);*/
				/*fc = fis.getChannel();
				mbb = fc.map(FileChannel.MapMode.READ_ONLY,datStart,datEnd-datStart+1);
				bb = mbb.asReadOnlyBuffer();*/
            } else {
                fis.skip(datStart - (6+4+8+8+8+8));
                
                /*data = new byte[datEnd-datStart+1];
                fis.read(data);*/
				/*fc = fis.getChannel();
				mbb = fc.map(FileChannel.MapMode.READ_ONLY,datStart,datEnd-datStart+1);
				bb = mbb.asReadOnlyBuffer();*/
                
                fis.skip(txtStart-datEnd-1);
                
                byte[] text = new byte[txtEnd-txtStart+1];
                fis.read(text);
                txt = new String(text);
            }
            
            Hashtable info = FCSTools.readText(txt);
            int byteorder;
            if(!info.containsKey("BYTEORD"))
                throw new ReadException("File Format Incorrect:\nno byte order specified");
            if(info.get("BYTEORD").equals("4,3,2,1"))
                byteorder = 1;
            else if(info.get("BYTEORD").equals("1,2,3,4"))
                byteorder = 2;
            else if(info.get("BYTEORD").equals("3,4,1,2"))
                byteorder = 3;
            else
                throw new ReadException("Unknown byte order -- file likely corrupted.");
            
            int datatype;
            if(!info.containsKey("DATATYPE"))
                throw new ReadException("File Format Incorrect:\nno data type specified");
            if(info.get("DATATYPE").equals("I")) //integer
                datatype = 1;
            else if(info.get("DATATYPE").equals("F")) //float
                datatype = 2;
            else if(info.get("DATATYPE").equals("D")) //double
                datatype = 3;
            else if(info.get("DATATYPE").equals("A")) //ASCII
                datatype = 4;
            else
                throw new ReadException("File Format Incorrect:\ndata type unrecognized");
            //if(datatype != 1)
            //    throw new ReadException("Non-integer data not supported");
            
            int mode;
            if(!info.containsKey("MODE"))
                throw new ReadException("File Format Incorrect:\nstorage mode not specified");
            if(info.get("MODE").equals("L")) //list
                mode = 1;
            else if(info.get("MODE").equals("C")) //correlated
                mode = 2;
            else if(info.get("MODE").equals("U")) //uncorr.
                mode = 3;
            else
                throw new ReadException("File Format Incorrect:\nmode unrecognized");
            if(mode != 1)
                throw new ReadException("Only list data supported.");
            
            int nParam;
            if(!info.containsKey("PAR"))
                throw new ReadException("File Format Incorrect:\nparameters unspecified");
            Integer I = new Integer((String)info.get("PAR"));
            nParam = I.intValue();
            
            //set up parameters
            FParameter[] params = new FParameter[nParam];
            int lenSum = 0;
			double[] logscale = new double[params.length]; //only used if data needs to be lin->log converted
			double[] logscale2 = new double[params.length];
            for(int i = 1; i <= params.length; i++) { //params start at #1
                if(!info.containsKey("P"+i+"B")) //bit length
                    throw new ReadException("File Format Incorrect:\nparameter "+i+" incomplete");
                I = new Integer((String)info.get("P"+i+"B"));
                int pb = I.intValue();
                if(pb % 8 != 0)
                    throw new ReadException("Only byte-justified data supported");
                pb = pb/8;
                lenSum += pb;
                
                if(!info.containsKey("P"+i+"R")) //range
                    throw new ReadException("File Format Incorrect:\nparameter "+i+" incomplete");
                I = new Integer((String)info.get("P"+i+"R"));
                int pe = I.intValue();
                
                if(!info.containsKey("P"+i+"E")) //amplification
                	info.put("P"+i+"E","0,0");
                    //throw new ReadException("File Format Incorrect:\nparameter "+i+" incomplete");
                int comma = ((String)info.get("P"+i+"E")).indexOf(',');
                String Spr1 = ((String)info.get("P"+i+"E")).substring(0,comma);
                String Spr2 = ((String)info.get("P"+i+"E")).substring(comma+1);
                Double d = new Double(Spr1); //was Integer for FCS2.0...
                int pr1 = d.intValue();
                d = new Double(Spr2);
                double pr2 = d.doubleValue();
                
                String ps;
                if(info.containsKey("P"+i+"S"))
                    ps = (String)info.get("P"+i+"S"); //long name
                else
                    ps = "";
                
                String pn;
                
                if(info.containsKey("P"+i+"N"))
                    pn = (String)info.get("P"+i+"N"); //short
                else
                    pn = "P"+i;
                
                params[i-1] = new FParameter(pb, pe, pr1, pr2, ps, pn);
			}
            
            //we're going to normalize all parameter ranges to one value
            int paramRange = 0;
            for(int i = 0; i < params.length; i++) {
            	if(params[i].range > paramRange)
            		paramRange = params[i].range;
            }
            for(int i = 0; i < params.length; i++)
            	params[i].range = paramRange;
            
            //log scale parameters
            for(int i = 0; i < params.length; i++) {
            	if(info.containsKey("P"+(i+1)+"DISPLAY")) { //numbers start from 1
					String islog = (String)info.get("P"+(i+1)+"DISPLAY");
					if(islog.equalsIgnoreCase("log")) {
						params[i].log = true;
						params[i].amp1 = (int)Math.ceil(Math.log10((double)params[i].range));
					}
				}
				if(!params[i].isLog && params[i].log) { //need to make logarithmic
					logscale[i] = (double)params[i].range/Math.ceil(Math.log10((double)params[i].range));
					logscale2[i] = logscale[i];
				} else if(params[i].isLog) {
					logscale[i] = (double)params[i].amp1/(double)params[i].range;
					logscale2[i] = (double)params[i].range/Math.ceil((double)params[i].amp1); //Math.log10((double)params[i].range)
				}
            }
            
            //correct data start/end positions if necessary (we've seen FlowJo output a funky file)
			int calcDataLen = 0;
		    for(int p = 0; p < params.length; p++) {
		    	calcDataLen += params[p].len;
		    }
		    calcDataLen *= Integer.parseInt((String)info.get("TOT"));
		    if((datEnd+1 - datStart < calcDataLen)) {
		    	if((datStart+calcDataLen <= f.length())) {
		    		datEnd = datStart+calcDataLen-1;
			    } else {
			    	throw new ReadException("File size does not match expected value!");
			    }
		  	}
		    
		    fc = fis.getChannel();
			mbb = fc.map(FileChannel.MapMode.READ_ONLY,datStart,datEnd-datStart+1);
			bb = mbb.asReadOnlyBuffer();
            
            //extract data
            //ByteArrayInputStream bios = new ByteArrayInputStream(data);
            int nCells = (datEnd-datStart+1)/lenSum; //data.length/(lenSum);
            if(nEvents != -1 && nEvents < nCells)
            	nCells = nEvents;
            int[][] cells = new int[nCells+1][nParam];
            int i = 0;
			double[] lin;
            while(bb.hasRemaining()) { //bios.available() > 0) {
            	if(nEvents != -1 && i > nEvents)
            		break;
            	
                cells[i] = new int[nParam];
				lin = new double[nParam];
                for(int j = 0; j < nParam; j++) {
                    /*byte[] p = new byte[params[j].len];
                    bb.get(p); //bios.read(p);*/
                    cells[i][j] = params[j].read(bb, byteorder, datatype);
					if(!params[j].isLog && params[j].log) { //need to make logarithmic
						lin[j] = (double)cells[i][j];
						cells[i][j] = (int)(Math.log10((double)cells[i][j])*logscale[j]);
					} else if(params[j].isLog) {
						lin[j] = Math.pow(10d,cells[i][j]*logscale[j]);
					} else {
						lin[j] = cells[i][j];
					}
                }
				if(comp) {
					lin = multiply(c,lin);
					for(int p = 0; p < params.length; p++) {
						if((!params[p].isLog && params[p].log) || params[p].isLog) {
							if(lin[p] <= 0.0) //prevent -infinity logs
								cells[i][p] = 0;
							else
								cells[i][p] = (int)(Math.log10(lin[p])*logscale2[p]);
						} else {
							cells[i][p] = (int)lin[p];
						}
					}
				}
                i++;
            }
			//System.out.println(System.currentTimeMillis());
            return new FACSData(dataIndex, f, null, info, params, cells, comp);
        } catch (IOException ioe) {
            throw new ReadException(ioe.toString());
        } catch (NumberFormatException nfe) {
        	nfe.printStackTrace();
            throw new ReadException("File Corrupted - numbers unreadable");
        }
    }

	public double[] multiply(double[][] comp, double[] cell) { //perform matrix*vector multiplication
		if(comp[0].length != cell.length) //matrix wrong size
			return cell;
		double[] out = new double[cell.length];
		for(int i = 0; i < cell.length; i++) {
			out[i] = dotProd(comp[i],cell);
		}
		return out;
	}

	public double dotProd(double[] one, double[] two) { //dot product of two vectors (as part of matrix mult)
		double out = 0;
		for(int i = 0; i < one.length; i++) {
			out += one[i]*two[i];
		}
		return out;
	}
    
    public int getInt(byte[] b) throws NumberFormatException {
        String s = new String(b);
        Integer i = new Integer(s.trim());
        return i.intValue();
    }
    
    /*public Hashtable readText(String txt) { //get key-value pairs out of list; keys have "$" at beginning; all are separated by a delimiter
        String del = txt.substring(0,1); //delimiter
        Matcher non = Pattern.compile("[^A-Za-z]").matcher(del);
        String del1 = del;
        if(non.find())
            del1 = "\\"+del;
        String[] list = txt.split(del1);
        
        Hashtable hash = new Hashtable();
        String key = "";
        for(int i = 0; i < list.length; i++) {
            if(i == 0 && list[i].equals(""))
                i++;
            if(list[i].equals("")) {
                String t = (String)hash.get(key);
                i++;
                hash.put(key,t+del+list[i]);
            } else {
                if(list[i].startsWith("$"))
                    key = list[i].substring(1);
                else
                    key = list[i];
                i++;
                hash.put(key,list[i]);
            }
        }
        return hash;
    }*/

	public void paintPlot(Graphics2D g2, Vector fGated, int param1, int param2, int mode, int resX, int resY, boolean hSecondMax, boolean cShowOutliers, boolean smooth) {
		//int param1 = this.param1, param2 = this.param2, mode = this.mode;

		int[][][] data = new int[fGated.size()][][];
		FParameter[][] params = new FParameter[fGated.size()][];
        for(int i = 0; i < fGated.size(); i++) {
            data[i] = ((FACSData)fGated.elementAt(i)).data;
			params[i] = ((FACSData)fGated.elementAt(i)).params;
        }

		Color[] colors = new Color[fGated.size()];
        colors[0] = Color.black;
        for(int i = 1; i < fGated.size(); i++) {
            colors[i] = Color.getHSBColor((float)i/(float)fGated.size(),1.0f,1.0f);
        }

		float max1, max2;

		g2.setColor(Color.white);
        g2.fillRect(0,0,resX,resY);

        for(int d = 0; d < data.length; d++) { //will only be multiple for HIST and DOTPLOT
            
            if(params[d].length != params[0].length)
                continue;
            if(!(mode == DOTPLOT || mode == HIST || mode == SM_HIST) && d > 0) //just in case...
                break;
            
            //generate plot points
            int[][] pts = data[d];
            
            /*boolean log1 = params[0][param1].log;
            boolean log2;
            if(param2 != -1)
                log2 = params[0][param2].log;
            else
                log2 = false;*/
            
            max1 = (float)params[0][param1].range;
            if(param2 != -1) {
                max2 = (float)params[0][param2].range;
            } else { //HIST or SM_HIST
                pts = genHistogram(data[d],param1,params,resX);
				
                max2 = getMax(pts, hSecondMax);
                
                param1 = 0; //these are just the ones for this function
                param2 = 1;
                max1 = resX;
            }
            
            int[][] twoHist = null;
            int dMax = 0;
            if(mode == DENSITY || mode == CONTOUR) {
                twoHist = gen2DHist(params, param1, param2, data[d],resX,1);
                dMax = get2DMax(twoHist);
                
                if(mode == CONTOUR) {
                	int nLevels = (int)(Math.log(dMax)/Math.log(3))+1;
                    int[] levels = new int[nLevels];
                    for(int i = 0; i < nLevels; i++) {
                        //levels[i-1] = (int)(dMax/(Math.pow(10d,(double)i/(double)nLevels)));
                    	levels[i] = (int)Math.pow(3,i);
                    }
                    
                    pts = ContourPlotter.getContourPoints(twoHist,levels,cShowOutliers);
                } else if(mode == DENSITY) {
                	if(smooth)
                		twoHist = ContourPlotter.blur(twoHist);
                    //pts = ones(params, param1, param2, params[0][param1].range/resX);
                	pts = null;
                }
                
                param1 = 0;
                param2 = 1;
                
				max1 = resX; max2 = resY;
            }
            //actual rendering
            g2.setColor(colors[d]);
            float lastX = 0, lastY = 0;
            int n_pts;
            if(pts == null) //for density plot (this is kinda bootstrapping but hey)
            	n_pts = resX*resY;
            else
            	n_pts = pts.length;
            for(int i = 0; i < n_pts; i++) {
            	//for density plot, we don't have a matrix (too much memory), so we'll calculate x-pos and y-pos
            	//we're going to calculate x-position and y-position from our index
            	//we'll go across then down, so...
            	int x_pos, y_pos;
            	if(pts == null) {
            		x_pos = i % resY;
                	y_pos = (i-x_pos) / resY;
            	} else {
            		x_pos = pts[i][param1];
            		y_pos = pts[i][param2];
            	}
                float x = x_pos/max1*(float)resX;
                float y = (max2-y_pos)/max2*(float)resY;
                if(mode == DOTPLOT) {
                    g2.draw(new Line2D.Float(x,y,x,y));
                } else if(mode == HIST) {
                    g2.draw(new Line2D.Float(lastX,lastY,x,y));
                    lastX = x; lastY = y;
                } else if(mode == SM_HIST) {
                    g2.draw(new Line2D.Float(lastX,lastY,x,y));
                    lastX = x; lastY = y;
                    i += resX/64; //256->4
                } else if(mode == DENSITY) {
                    //int den = twoHist[pts[i][0]][pts[i][1]]; //density at cur pos
                    int den = twoHist[x_pos][y_pos];
                    
                    if(den != 0) {
                        float ratio = Math.min(1f, (float)den/(float)dMax);
                        Color c = Color.getHSBColor(0.25f+0.125f+ratio,1f,1f);
                        g2.setColor(c);
                        g2.draw(new Line2D.Float(x,y,x,y));
                    }
                } else if(mode == CONTOUR) {
					g2.draw(new Line2D.Float(x,y,x,y));
                }
            }
        }
	}

	public int[][] genHistogram(int[][] data, int param, FParameter[][] params, int resX) {
        int[][] h = new int[resX][2];
        int smooth = params[0][param].range/resX;
        for(int i = 0; i < resX; i++) {
            h[i][0] = i;
            h[i][1] = 0;
        }
        for(int i = 0; i < data.length; i++) {
        	int val1 = data[i][param]/smooth;
			if(val1 < 0)
				val1 = 0;
			if(val1 == resX)
				val1--;
            h[val1][1]++;
        }
        return h;
    }
    
    public int getMax(int[][] data, boolean hSecondMax) {
        int max = 0;
        int start = 0;
        if(hSecondMax) //okay, so it's not "second max", but it works...
        	start += 5;
        int end = data.length;
        if(hSecondMax)
        	end -= 5;
        for(int i = start; i < end; i++) {
            if(data[i][1] > max) {
                max = data[i][1];
            }
        }
        return max;
    }
    
    public int[][] ones(FParameter[][] params, int param1, int param2, int smooth) {
        //okay, so it's not really ones, but it contains every point in the 2D plot
        int[][] a = new int[params[0][param1].range/smooth*params[0][param2].range/smooth][2];
        
        int n = 0;
        for(int i = 0; i < params[0][param1].range/smooth; i++) {
            for(int j = 0; j < params[0][param2].range/smooth; j++) {
                a[n][0] = i;
                a[n][1] = j;
                n++;
            }
        }
        return a;
    }
    
    public int[][] gen2DHist(FParameter[][] params, int param1, int param2, int[][] data, int res, int smooth2) {
		int smooth_x = params[0][param1].range/res;
		int smooth_y = params[0][param2].range/res;
        int[][] h = new int[res][res];
        for(int i = 0; i < res; i++) {
            for(int j = 0; j < res; j++) {
                h[i][j] = 0;
            }
        }
        for(int i = 0; i < data.length; i++) {
            int val1 = (data[i][param1]/smooth_x)*1/smooth2;
			if(val1 < 0)
				val1 = 0;
            int val2 = (data[i][param2]/smooth_y)*1/smooth2;
			if(val2 < 0)
				val2 = 0;
            for(int j = 0; j < smooth2; j++) {
                for(int k = 0; k < smooth2; k++) {
                    h[val1*smooth2+j][val2*smooth2+k]++;
                }
            }
        }
        return h;
    }
    
    public int get2DMax(int[][] data) {
        int max = 0;
        int border = 5; //ignore axes
        for(int i = border; i < data.length-border; i++) {
            for(int j = border; j < data[i].length-border; j++) {
                if(data[i][j] > max) {
                    max = data[i][j];
                }
            }
        }
        return max;
    }

	/**
	 * This function will gate fData, editing the objects in place.
	 * 
	 * @param regions RVector containing all defined regions.
	 * @param fData Vector of FACSData objects to gate.
	 * @param gates Vector of Gate objects describing which gates to use.
	 */
	public void gate(RVector regions, Vector fData, Vector gates) {
        //FACSData[] out = new FACSData[in.length];
        
        for(int d = 0; d < fData.size(); d++) { //for each data
            //initialize shapes/params per shape
            Vector shapes = new Vector();
            Vector params = new Vector();
            
            FACSData f = (FACSData)fData.elementAt(d);
            /*shapes.add(new Rectangle2D.Double(0.0,0.0,f.params[0].range,f.params[0].range));
            double[] zz = { 0.0, 0.0 };
            params.add(zz);*/
			
            for(int i = 0; i < gates.size(); i++) { //for each gate
                Gate g = (Gate)gates.elementAt(i);
                double[] gate = ((Region)regions.elementAt(g.region)).getVals(f.filename);
                Shape s = null;
                if(arrayContains(g.data,f.filename))
                    s = getShape(gate,f.params);
                shapes.add(s);
                double[] gparams = { gate[1], gate[2] };
                params.add(gparams);
            }
            
			if(shapes.size() == 0) {
				fData.setElementAt(new FACSData(f.index, f.f, f.name, f.text, f.params, f.data, f.comp),d);
				continue;
			}
            
            Vector filtered = new Vector();
            for(int i = 0; i < f.data.length; i++) { //for each cell
                boolean okay = true;
                for(int g = 0; g < shapes.size(); g++) { //again for each gate
                    Shape s = (Shape)shapes.elementAt(g);
                    double[] gparams = (double[])params.elementAt(g);
                    if(s == null)
                        continue;
                    if(!s.contains(f.data[i][(int)gparams[0]],f.data[i][(int)gparams[1]])) {
                        okay = false;
                        break;
                    }
                }
                if(okay)
                    filtered.add(f.data[i]);
            }
            fData.setElementAt(new FACSData(f.index, f.f, f.name, f.text, f.params, makeArray(filtered,f.params.length), f.comp),d);
        }
        
        //return out;
    }

	public static boolean arrayContains(String[] array, String item) {
        for(int i = 0; i < array.length; i++) {
            if(array[i].equals(item))
                return true;
        }
        return false;
    }

	public int[][] makeArray(Vector in,int nParams) {
        int[][] out = new int[in.size()][nParams];
        for(int i = 0; i < in.size(); i++) {
            out[i] = (int[])in.elementAt(i);
        }
        return out;
    }
    
    public static int[] makeArray(Vector in) {
        int[] out = new int[in.size()];
        for(int i = 0; i < in.size(); i++) {
            out[i] = ((Integer)in.elementAt(i)).intValue();
        }
        return out;
    }
	public Shape getShape(double[] gate, FParameter[] params) {
        Shape s = null;
        if((int)gate[0] == RECT) {
            s = new Rectangle2D.Double(gate[3],gate[4],gate[5],gate[6]);
        } else if((int)gate[0] == ELLIPSE) {
            Ellipse2D.Double el = new Ellipse2D.Double(0,-gate[6]/2,gate[5],gate[6]);
            AffineTransform atEl = AffineTransform.getTranslateInstance(gate[3],gate[4]+gate[6]/2);
            atEl.rotate(-gate[7]);
            s = atEl.createTransformedShape(el);
        } else if((int)gate[0] == RANGE) {
            double theta = Math.atan((gate[6]-gate[4])/(gate[5]-gate[3]));
            if((theta < Math.PI/5 && theta > -Math.PI/5) || gate[2] == -1)
                s = new Rectangle2D.Double(gate[3],0.0,gate[5]-gate[3],params[(int)gate[1]].range); //horiz
            else
                s = new Rectangle2D.Double(0.0,gate[6],params[(int)gate[2]].range,gate[4]-gate[6]);
        } else if((int)gate[0] == POLY) {
            int[] xp = new int[(gate.length-3)/2];
            int[] yp = new int[(gate.length-3)/2];
            int n = 0;
            for(int p = 3; p < gate.length; p++) {
                xp[n] = (int)gate[p];
                p++;
                yp[n] = (int)gate[p];
                n++;
            }
            s = new Polygon(xp,yp,(gate.length-3)/2);
        }
        return s;
    }
}

class Gate {
    //int[] dataIndices;
	String[] data;
    int region;
    String name;
    
    public Gate(String[] d, int r, String name) {
        data= d;
        region = r;
        this.name = name;
    }
    
    public String toString() {
        return name;
    }
    
    public String getWritable() {
        String w = region+"|";
        for(int i = 0; i < data.length; i++) {
            w += data[i]+"|";
        }
        return w.substring(0,w.length()-1);
    }
    
    public static Gate fromWritable(String s, String name) {
        String[] pcs = s.split("\\|");
        int region = new Integer(pcs[0]).intValue();
        String[] di = new String[pcs.length-1];
        for(int i = 1; i < pcs.length; i++) {
            //di[i-1] = new Integer(pcs[i]).intValue();
			di[i-1] = pcs[i];
        }
        return new Gate(di, region, name);
    }
}